# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2011 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
`filestore` - dmedia hashing protocol and file layout.

This is the heart of dmedia.  It's been split out of dmedia to make it easier
to review, and in hopes that other apps might use the hashing protocol and
`FileStore` even if they don't use the higher level dmedia features.

Learn more about dmedia (aka the Distributed Media Library) here:

    https://launchpad.net/dmedia



Request For Comment
===================

We're seeking feedback from the crypto community on the dmedia hashing protocol
before declaring it final.

In particular, we'd love your thoughts on:

    1. Potential drawbacks to using a hash-list

    2. How we cryptographically tie the root-hash to the file-size

    3. Whether the 240-bit Skein-512 digest size is too small, too large, or
       just right

    4. Whether tying the leaf-hash to the leaf-index (leaf position) is useful
       enough to justify the (small) increase in protocol complexity



The Hashing Protocol
====================

Files in dmedia are given a globally unique ID based on their content-hash
(pretty straightforward).  However, the hashing protocol is designed to solve
some key problems faced when transferring large files over unreliable (if not
untrusted) networks.  Any content-hash is a self-authenticating name once you
have the entire file, but getting that entire file in the first place is the
tricky part.  We need the content-hash also to self-authenticate:

    * The file-size before we even start downloading
    * Chunks of the file at a reasonably small granularity as we download

The dmedia content-hash is calculated using a hash-list with 8 MiB leaves.  The
root-hash is cryptographically tied to the file-size, and the leaf-hashes are
cryptographically tied to the leaf-index (the position of the leaf in the file,
starting from zero).

In a nutshell, this is the algorithm:

    ::

        _id = hash(
            size + hash(0 + leaf_0) + hash(1 + leaf_1) ... + hash(N + leaf_N)
        )

Where integers like ``size`` and ``0`` are hashed according to their UTF-8
encoded decimal representation.

As dmedia stores both the file-size and the leaf-hashes in a file's CouchDB
document, the protocol gives you a cryptographically strong way to verify that
for a given root-hash (_id):

    1) the claimed file-size is correct
    2) the claimed leaf-hashes are correct
    3) the claimed data for any leaf is correct

dmedia uses the Skein hash, a fast, flexible, utterly pragmatic cryptographic
hash function (and it's one of the 5 finalist in the SHA-3 competition):

    http://www.skein-hash.info/

Specifically, dmedia uses the excellent PySkein implementation by Hagen
Fürstenau:

    http://packages.python.org/pyskein/

There are two distinct types of hashes involved in the protocol:

    1. `hash_leaf()` hashes a leaf, tying it to the leaf's position in the file

    2. `hash_root()` hashes the concatenated leaf-hashes, tying it to the file
       size

As per recommendation in the Skein specification, a distinct personalization
string is used for each type of hash:

    ::

        20110430 jderose@novacut.com dmedia/leaf
        20110430 jderose@novacut.com dmedia/root

As far as why one should use such personalization strings, here is what the
Skein specification has to say on the matter:

    All Skein applications (except the PRNG output production) can be
    personalized with the personalization input.  We recommend that all
    application designers seriously consider doing this; we have seen many
    protocols where a hash that is computed in one part of the protocol can be
    used in an entirely different part because two hash computations were done
    on similar or related data, and the attacker can force the application to
    make the hash inputs the same.  Personalizing each hash function used in
    the protocol summarily stops this type of attack.

For further details see section 4.11 in the Skein specification:

    http://www.skein-hash.info/sites/default/files/skein1.3.pdf

The `hash_root()` and `hash_leaf()` functions provide high-level building
blocks.  For example, you calculate the dmedia hash of the 22-byte file
containing the UTF-8 text "Hello, Bruce Schneier!" like this:

>>> hash_root(22, hash_leaf(0, b'Hello, Bruce Schneier!'))
'V3ODQNYR62VFR4KGIKS3IE4SQKCYI7GIYLPD3UA4RDSCR7HG'


Design Decision: base32-encoded 240-bit content-hash
----------------------------------------------------

The `FileStore` layout was designed to allow the canonical filename to be
constructed from the content-hash in the simplest way possible, without
requiring any special decoding or encoding.  For this reason, the content-hash
(as stored in CouchDB) is base32-encoded.

Base32-encoding was chosen because:

    1. It's more compact than base16/hex

    2. It can be used to name files on case *insensitive* filesystems (whereas
       base64-encoding cannot)

To be "cleanly" base32-encoded (that is, not require padding), the input must
be a multiple of 40-bits (5-bytes).  dmedia is all about being as simple as
possible, so that's why only digest sizes that are a multiple of 40 bits were
up for consideration.

As Skein specifically supports arbitrary digest sizes, this is a clear win for
simplicity.  Although base32-encoding is only a concern for the root-hash, the
hashing protocol uses the same digest-size throughout, just to keep things as
simple as possible.

Although it was not a requirement, a 240-bit digest has the added benefit of
also being "cleanly" base64-encoded.


Design Decision: fixed-sized leaves
-----------------------------------

It's tempting to make the leaf-size a function of the file-size, but there is a
serious drawback in that you couldn't compute a hash as a file is being written,
before the ultimate file-size is know.

A key design consideration for the hashing protocol was to allow capture devices
(cameras, recorders, etc) to write files directly into a dmedia-like filesystem.
dmedia may or may not find a home in consumer electronics, but it's an important
consideration nonetheless.


Design Decision: cryptographically tie root-hash to file-size
-------------------------------------------------------------

When dmedia downloads a file, the first question it must answer is, "is there
enough space for the file?"  And if there is enough space, dmedia will
pre-allocate the file using `fallocate`, and then start downloading.

If malicious peers could easily lie about the file-size, a lot of havoc could be
wreaked by, say, causing ridiculously large files to be pre-allocated.  Although
as dmedia uses a hash-list with fixed sized leaves, the client *could* validate
the file-size to within (LEAF_SIZE - 1) bytes.

But what we really want is to be able to validate the *exact* file-size.  To do
this we use the UTF-8 encoded string representation of the file-size (base10)
as the first piece of data to hash:

>>> file_size = 31415
>>> h = skein512(str(file_size).encode('utf-8'),
...     digest_bits=240,
...     pers=b'20110430 jderose@novacut.com dmedia/root',
... )

And then we update the hash with the concatenated leaf-hashes:

>>> h.update(b'The Concatenated Leaf Hashes')
>>> b32encode(h.digest())
b'4DSH342K5YAXRJKOFXBFBIT4HFZUGY2W3XX5ZVACLTS55HBP'


Design Decision: cryptographically tie leaf-hash to leaf-index
--------------------------------------------------------------

The leaf-hash is cryptographically tied to the position of the leaf within the
file, using the leaf-index (starting from zero).

To do this, we use the UTF-8 encoded string representation of the leaf-index
(base10) as the first piece of data to hash:

>>> leaf_index = 2
>>> h = skein512(str(leaf_index).encode('utf-8'),
...     digest_bits=240,
...     pers=b'20110430 jderose@novacut.com dmedia/leaf',
... )

And then we update the hash with leaf data, like this:

>>> h.update(b'The Leaf Data')
>>> b32encode(h.digest())
b'YELUPIBCMBD3AAO34D3KJFO2M6ALBAF2NQEOSDXBSY5KVBAN'

The motivation for this is to force implementations to do the right thing.  In
particular it forces peers to stay in lockstep agreement about which leaf
(positionally) is being transfered.

When this feature was added, we noticed it made our reference implementation
clearer and simpler as code dealing with a leaf must always know the leaf-index,
something which is generally needed anyway (say when reading a specific leaf
from a file).

We'd like feedback on this aspect of the protocol, but it has proven beneficial
in our reference implementation.  If no good arguments are raised against this,
it will remain in the final protocol.


Design Decision: migration strategy is to change digest-size
------------------------------------------------------------

dmedia has been very deliberately designed to have just "one way of doing it".
We toyed with idea of allow multiple hash algorithms, digest sizes, etc.  But
the problem is options like these are an enormous burden on those implementing
the protocol, are big barrier to adoption.  To have a chance at becoming
ubiquitous in the consumer electronics space especially, dmedia must be
painfully simple.

So once the hashing protocol is finalized, it will be the one and only way that
dmedia content-hashes are calculated... for the foreseeable future (to the
extend this possible in cryptography).

However, the unforeseeable might happen, so it's important to have a clear plan
for how we would migrate to a different hashing protocol.  And that plan is, if
the hash needs to change, it will have different digest length.  This makes it
easy to tell if a particular ID was computed using the v1 or v2 hashing
protocol.



The FileStore
=============

To create a `FileStore`, pass it the parent directory the store will be
contained in.  For example:

>>> parent = tempfile.mkdtemp()
>>> fs = FileStore(parent)

The `FileStore` will reside in a ``'.dmedia'`` directory inside of the parent
directory:

>>> path.isdir(path.join(parent, '.dmedia'))
True
>>> fs.base == path.join(parent, '.dmedia')
True

There will be a number of directories inside the ``'.dmedia'`` control
directory:

>>> sorted(os.listdir(fs.base))
['README.txt', 'corrupt', 'files', 'partial', 'tmp']


"""

import os
from os import path
import io
import tempfile
import stat
from subprocess import check_call, CalledProcessError
from base64 import b32encode, b32decode
from threading import Thread
from queue import Queue
from collections import namedtuple
import logging

from skein import skein512


__version__ = '11.09.0'

log = logging.getLogger('filestore')

# Constants defined by the dmedia hashing protocol:
PERS_LEAF = b'20110430 jderose@novacut.com dmedia/leaf'
PERS_ROOT = b'20110430 jderose@novacut.com dmedia/root'
DIGEST_BITS = 240
DIGEST_BYTES = DIGEST_BITS // 8
LEAF_SIZE = 8 * 1024 * 1024  # 8 MiB leaf size

# Handy constants for file layout:
FALLOCATE = '/usr/bin/fallocate'
DOTNAME = '.dmedia'
B32ALPHABET = '234567ABCDEFGHIJKLMNOPQRSTUVWXYZ'
B32NAMES = tuple(a + b for a in B32ALPHABET for b in B32ALPHABET)

# Some handy namedtuples:
ContentHash = namedtuple('ContentHash', 'id file_size leaf_hashes')
File = namedtuple('File', 'name size mtime')
Leaf = namedtuple('Leaf', 'index data')

# Provide very clear TypeError messages:
TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'

# For the README.txt in .dmedia directories:
README = """This directory is automatically managed by dmedia.

Warning: if you make manual changes in this directory, it will quite likely
result in data loss!

This is a control directory for the Distributed Media Library (aka dmedia).  It
contains files and directories arranged in the special FileStore layout.

To learn more about dmedia, see:

    https://launchpad.net/dmedia
"""


class RootHashError(Exception):
    """
    Raised when file_size and/or leaf_hashes are invalid.

    This is exception has ``id``, ``file_size``, and ``leaf_hashes`` attributes,
    and a ``bad_id`` attribute with the root-hash computed from the supplied
    file_size and leaf_hashes.
    """
    def __init__(self, _id, file_size, leaf_hashes, bad_id):
        self.id = _id
        self.file_size = file_size
        self.leaf_hashes = leaf_hashes
        self.bad_id = bad_id
        super().__init__('{!r}'.format(_id))


class NestedParent(Exception):
    """
    Raised when a `FileStore` parent directory is nested in another filestore.
    """
    def __init__(self, parent, dotname):
        self.parent = parent
        self.dotname = dotname
        super().__init__('{!r} contains {!r}'.format(parent, dotname))


class AmbiguousPath(Exception):
    """
    Raised when a path is relative or non-normalized.
    """
    def __init__(self, pathname, abspath):
        self.pathname = pathname
        self.abspath = abspath
        super().__init__('{!r} resolves to {!r}'.format(pathname, abspath))


class PathTraversal(Exception):
    """
    Raised when a path is outside of a `FileStore` base directory.
    """
    def __init__(self, base, pathname, abspath):
        self.base = base
        self.pathname = pathname
        self.abspath = abspath
        super().__init__('{!r} outside of {!r}'.format(abspath, base))


class IntegrityError(Exception):
    """
    Base class for various file integrity errors.

    All `IntegrityError` subclasses will have 2 important attributes that your
    error handling code may need to use:

        * ``parent`` - the parent directory in which the `FileStore` resides,
          in case it is unclear which `FileStore`

        * ``id`` - the ID (root-hash) of the corrupt file
    """
    def __init__(self, parent, _id):
        self.parent = parent
        self.id = _id
        super().__init__('{!r} is corrupted in {!r}'.format(_id, parent))


class FileIntegrityError(IntegrityError):
    """
    Raised if a file does not have the expected root-hash.

    This exception means the entire file was hashed and the result did not equal
    the file ID.
    """
    def __init__(self, parent, _id, bad_id):
        self.bad_id = bad_id
        super().__init__(parent, _id)


class LeafIntegrityError(IntegrityError):
    """
    Raised if a leaf does not have the expected leaf-hash.

    This exception means that during an incremental verification, a corrupted
    leaf was encountered.
    """
    def __init__(self, parent, _id, leaf_index, leaf_hash, bad_leaf_hash):
        self.leaf_index = leaf_index
        self.leaf_hash = leaf_hash
        self.bad_leaf_hash = bad_leaf_hash
        super().__init__(parent, _id)


class SizeIntegrityError(IntegrityError):
    """
    Raised when the on-disk file size doesn't match up with the root-hash.

    The cheapest verification is to simply check the file-size against what
    the root-hash dictates.

    To ensure that the claimed file-size and leaf-hashes are correct for the
    root-hash, check them with `check_root_hash()` first.
    """
    def __init__(self, parent, _id, file_size, bad_file_size):
        self.file_size = file_size
        self.bad_file_size = bad_file_size
        super().__init__(parent, _id)



def hash_leaf(leaf_index, leaf_data=None):
    """
    Return the leaf-hash of the leaf at (zero) index *leaf_index*.

    If you only supply *leaf_index*, and appropriately initialized ``skein512``
    object is returned:

    >>> hash_leaf(2)  #doctest: +ELLIPSIS
    <Skein-512 hash object at 0x...>

    If you also supply the *leaf_data*, the 240-bit digest is returned:

    >>> b32encode(hash_leaf(2, b'XYZ'))
    b'DHS36R6HGG52TQRB6BN7N66EWOPU4H67WISD3J6TWNCXXAZT'

    :param leaf_index: an ``int`` >= 0
    :param leaf_data: optional ``bytes`` instance with contents of this leaf
    """
    if not isinstance(leaf_index, int):
        raise TypeError(
            TYPE_ERROR.format('leaf_index', int, type(leaf_index), leaf_index)
        )
    if leaf_index < 0:
        raise ValueError(
            'leaf_index: must be >= 0; got {!r}'.format(leaf_index)
        )
    lh = skein512(str(leaf_index).encode('utf-8'),
        digest_bits=DIGEST_BITS,
        pers=PERS_LEAF,
    )
    if leaf_data is None:
        return lh
    if not isinstance(leaf_data, bytes):
        raise TypeError(
            TYPE_ERROR.format('leaf_data', bytes, type(leaf_data), leaf_data)
        )
    if not (1 <= len(leaf_data) <= LEAF_SIZE):
        raise ValueError('Need 1 <= len(leaf_data) <= {}; got {}'.format(
                LEAF_SIZE, len(leaf_data)
            )
        )
    lh.update(leaf_data)
    return lh.digest()


def hash_root(file_size, leaf_hashes):
    """
    Return the root-hash for a file *file_size* bytes with *leaf_hashes*.

    For example:

    >>> hash_root(31415, b'NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN')
    'OMPLTTYVTIJINDZWIS2PBZ4THWA6CTGCGT27RFIDKV7FSTCA'

    :param file_size: an ``int`` >= 1
    :param leaf_hashes: a ``bytes`` instance that is the concatenated leaf
        hashes produced by `hash_leaf()`
    """
    if not isinstance(file_size, int):
        raise TypeError(
            TYPE_ERROR.format('file_size', int, type(file_size), file_size)
        )
    if file_size < 1:
        raise ValueError(
            'file_size: must be >= 1; got {!r}'.format(file_size)
        )
    check_leaf_hashes(leaf_hashes)
    n = len(leaf_hashes) // DIGEST_BYTES
    low = (n - 1) * LEAF_SIZE + 1
    high = n * LEAF_SIZE
    if not (low <= file_size <= high):
        raise ValueError('file_size is {}; need {} <= file_size <= {}'.format(
                file_size, low, high
            )
        )
    h = skein512(str(file_size).encode('utf-8'),
        digest_bits=DIGEST_BITS,
        pers=PERS_ROOT,
    )
    h.update(leaf_hashes)
    return b32encode(h.digest()).decode('utf-8')


def check_leaf_hashes(leaf_hashes):
    """
    Verify that *leaf_hashes* is a ``bytes`` instance of correct length.

    If *leaf_hashes* is valid, it is returned unchanged:

    >>> check_leaf_hashes(b'Y' * 30)
    b'YYYYYYYYYYYYYYYYYYYYYYYYYYYYYY'

    If *leaf_hashes* is invalid, a ``TypeError`` or ``ValueError`` is raised:

    >>> check_leaf_hashes(b'N' * 15)
    Traceback (most recent call last):
      ...
    ValueError: len(leaf_hashes) must be multiple of 30; got 15: b'NNNNNNNNNNNNNNN'

    """
    if not isinstance(leaf_hashes, bytes):
        raise TypeError(TYPE_ERROR.format(
                'leaf_hashes', bytes, type(leaf_hashes), leaf_hashes
            )
        )
    if len(leaf_hashes) == 0:
        raise ValueError('leaf_hashes cannot be empty')
    if len(leaf_hashes) % DIGEST_BYTES != 0:
        raise ValueError(
            'len(leaf_hashes) must be multiple of {}; got {}: {!r}'.format(
                DIGEST_BYTES, len(leaf_hashes), leaf_hashes
            )
        )
    return leaf_hashes


def check_root_hash(_id, file_size, leaf_hashes):
    """
    Verify that claimed *file_size* and *leaf_hashes* are correct for *_id*.

    For example, if the *file_size* was tampered with:

    >>> leaf_hash = hash_leaf(0, b'Hello, Bruce Schneier!')
    >>> _id = hash_root(22, leaf_hash)
    >>> check_root_hash(_id, 21, leaf_hash)  # 21 instead of 22 bytes
    Traceback (most recent call last):
      ...
    filestore.RootHashError: 'V3ODQNYR62VFR4KGIKS3IE4SQKCYI7GIYLPD3UA4RDSCR7HG'

    If the claimed *file_size* and *leaf_hashes* are correct, the 3 values are
    returned in a `ContentHash` named tuple:

    >>> ch = check_root_hash(_id, 22, leaf_hash)
    >>> ch.id
    'V3ODQNYR62VFR4KGIKS3IE4SQKCYI7GIYLPD3UA4RDSCR7HG'
    >>> ch.file_size
    22

    """
    computed = hash_root(file_size, leaf_hashes)
    if _id != computed:
        raise RootHashError(_id, file_size, leaf_hashes, computed)
    return ContentHash(_id, file_size, leaf_hashes)


def enumerate_leaf_hashes(leaf_hashes):
    """
    Enumerate *leaf_hashes*.

    This will yield ``(leaf_index,leaf_hash)`` tuples for each leaf-hash in
    *leaf_hashes* For example:

    >>> leaf_hashes = b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBBB'
    >>> list(enumerate_leaf_hashes(leaf_hashes))
    [(0, b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'), (1, b'BBBBBBBBBBBBBBBBBBBBBBBBBBBBBB')]

    """
    check_leaf_hashes(leaf_hashes)
    for i in range(len(leaf_hashes) // DIGEST_BYTES):
        yield (i, leaf_hashes[i*DIGEST_BYTES : (i+1)*DIGEST_BYTES])


def get_leaf_hash(leaf_hashes, i):
    """
    Return a slice containing the i-th leaf_hash from *leaf_hashes*.

    For example:

    >>> lh = b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBBB'
    >>> get_leaf_hash(lh, 0)
    b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'
    >>> get_leaf_hash(lh, 1)
    b'BBBBBBBBBBBBBBBBBBBBBBBBBBBBBB'
    >>> get_leaf_hash(lh, 2)
    b''

    """
    return leaf_hashes[i*DIGEST_BYTES : (i+1)*DIGEST_BYTES]


def missing_leaves(fp, leaf_hashes):
    """
    Resume a transfer by iterating through leaves that are corrupted or missing.

    This will yield an ``(index,leaf_hash)`` tuple for each leaf that must be
    downloaded.
    """
    fp.seek(0)
    for (i, leaf_hash) in enumerate_leaf_hashes(leaf_hashes):
        data = fp.read(LEAF_SIZE)
        if not data or hash_leaf(i, data) != leaf_hash:
            yield (i, leaf_hash)


def safe_path(pathname):
    """
    Ensure that *pathname* is a normalized absolute path.

    This is to help protect against path-traversal attacks and to prevent use of
    ambiguous relative paths.

    For example, if *pathname* is not a normalized absolute path,
    `AmbiguousPath` is raised:

    >>> safe_path('/foo/../root')
    Traceback (most recent call last):
      ...
    filestore.AmbiguousPath: '/foo/../root' resolves to '/root'

    Otherwise *pathname* is returned unchanged:

    >>> safe_path('/foo/bar')
    '/foo/bar'

    """
    if path.abspath(pathname) != pathname:
        raise AmbiguousPath(pathname, path.abspath(pathname))
    return pathname


def safe_id(_id):
    """
    Verify that *_id* is a valid base32-encoded ID of the correct length.

    A malicious *_id* could cause path traversal or other security gotchas,
    thus this sanity check.  When *_id* is valid, it is returned unchanged:

    >>> safe_id('OMPLTTYVTIJINDZWIS2PBZ4THWA6CTGCGT27RFIDKV7FSTCA')
    'OMPLTTYVTIJINDZWIS2PBZ4THWA6CTGCGT27RFIDKV7FSTCA'

    However, when *_id* does not conform, a ``TypeError`` or ``ValueError`` is
    raised:

    >>> safe_id('NWBNVXVK5DQGIOW7MYR4K3KA')
    Traceback (most recent call last):
      ...
    ValueError: _id: decoded must be 30 bytes; got 15: 'NWBNVXVK5DQGIOW7MYR4K3KA'

    """
    if not isinstance(_id, str):
        raise TypeError(TYPE_ERROR.format('_id', str, type(_id), _id))
    try:
        bin_id = b32decode(_id.encode('ascii'))
    except (ValueError, TypeError) as e:
        raise ValueError('_id: cannot b32decode {!r}: {}'.format(_id, e))
    if len(bin_id) != DIGEST_BYTES:
        raise ValueError('_id: decoded must be {} bytes; got {}: {!r}'.format(
            DIGEST_BYTES, len(bin_id), _id)
        )
    return _id


class Hasher(object):
    """
    A helper to keep track of state as you hash leaf after leaf.
    """

    __slots__ = ('file_size', 'leaf_index', 'array', 'closed')

    def __init__(self):
        self.file_size = 0
        self.leaf_index = 0
        self.array = bytearray()
        self.closed = False

    def update(self, leaf):
        assert isinstance(leaf, bytes)
        assert 1 <= len(leaf) <= LEAF_SIZE
        assert not self.closed
        if len(leaf) < LEAF_SIZE:
            self.closed = True

        leaf_hash = hash_leaf(self.leaf_index, leaf)
        self.array.extend(leaf_hash)
        self.file_size += len(leaf)

        low = self.leaf_index * LEAF_SIZE + 1
        high = (self.leaf_index + 1) * LEAF_SIZE
        assert low <= self.file_size <= high

        self.leaf_index += 1
        return leaf_hash

    def content_hash(self):
        leaf_hashes = bytes(self.array)
        return ContentHash(
            hash_root(self.file_size, leaf_hashes),
            self.file_size,
            leaf_hashes
        )


def reader_worker(src_fp, q):
    """
    Simple reader worker used by `reader_iter()`.
    """
    assert isinstance(src_fp, (io.BufferedReader, io.BufferedRandom))
    assert src_fp.mode in ('rb', 'rb+')
    src_fp.seek(0)
    while True:
        leaf_data = src_fp.read(LEAF_SIZE)
        q.put(leaf_data)
        if not leaf_data:
            break


def reader_iter(src_fp):
    """
    Iterate through leaves in *src_fp*, reading in a separate thread.

    The function yields a ``(leaf_index, leaf_data)`` tuple for each leaf in
    *src_fp*.

    Many operations in the `FileStore` involve a loop where an 8 MiB leaf is
    read, then hashed, then possibly written to disk.  As the hashing
    performance is almost always IO bound on the read side, putting the reader
    in a separate thread can greatly improve throughput.

    On the consumer side, this function acts as a simple iterator, completely
    hiding the reader thread.  Typical usage will be something like this:

    >>> def example():
    ...     h = Hasher()
    ...     src_fp = open('/my/file', 'rb')
    ...     for (leaf_index, leaf) in reader_iter(src_fp):
    ...         h.update(leaf)
    ...

    The thread target is the `reader_worker()` function.

    To control memory usage, the queue is created with a 4 item maxsize, meaning
    at most 32 MiB will be queued.  This is large enough to maximize throughput
    in most cases, but small enough as to accommodate fairly svelte systems.
    """
    if not isinstance(src_fp, (io.BufferedReader, io.BufferedRandom)):
        raise TypeError(TYPE_ERROR.format('src_fp',
                (io.BufferedReader, io.BufferedRandom), type(src_fp), src_fp
            )
        )
    assert src_fp.mode in ('rb', 'rb+')
    q = Queue(4)
    thread = Thread(target=reader_worker, args=(src_fp, q))
    thread.daemon = True
    thread.start()
    i = 0
    while True:
        leaf = q.get()
        if not leaf:
            break
        yield (i, leaf)
        i += 1
    thread.join()  # Make sure reader_worker() terminates


def hash_fp(src_fp, dst_fp=None):
    """
    Compute content hash of open file *src_fp*, optionally writing *dst_fp*.

    This function uses `reader_iter()` to transparently read from *src_fp* in
    another thread.

    The return value is a `ContentHash` named tuple with the ``id``,
    ``file_size``, and ``leaf_hashes``.

    :param src_fp: a file opened in ``'rb'`` or ``'rb+'`` mode
    :param dst_fp: optional file opened in ``'wb'`` mode
    """
    hasher = Hasher()
    for (i, leaf) in reader_iter(src_fp):
        hasher.update(leaf)
        if dst_fp:
            dst_fp.write(leaf)
    return hasher.content_hash()


def fallocate(size, filename):
    """
    Attempt to efficiently preallocate file *filename* to *size* bytes.

    If the fallocate command is available, it will always at least create an
    empty file (the equivalent of ``touch filename``), even the file-system
    doesn't support pre-allocation.
    """
    if not isinstance(size, int):
        raise TypeError(TYPE_ERROR.format('size', int, type(size), size))
    if size <= 0:
        raise ValueError('size must be >0; got {!r}'.format(size))
    filename = safe_path(filename)
    if not path.isfile(FALLOCATE):
        return None
    try:
        check_call([FALLOCATE, '-l', str(size), filename])
        return True
    except CalledProcessError:
        return False


def ensuredir(d):
    """
    Ensure that *d* exists, is a directory, and is not a symlink.
    """
    try:
        os.mkdir(d)
        return True
    except OSError:
        if not path.isdir(d):
            raise ValueError('not a directory: {!r}'.format(d))
        if path.islink(d):
            raise ValueError(
                '{!r} is symlink to {!r}'.format(d, os.readlink(d))
            )
        return False


def iter_files(base):
    """
    Recursively iterate through files in directory *base*.

    Note that this does not catch errors like ``OSError``.  We specifically want
    these errors to propagate up!  We don't want a permission error to be
    interpreted as there being no files on the card!
    """
    dirs = []
    for name in sorted(os.listdir(base)):
        fullname = path.join(base, name)
        s = os.lstat(fullname)
        if stat.S_ISLNK(s.st_mode):
            continue
        if stat.S_ISREG(s.st_mode):
            yield File(fullname, s.st_size, s.st_mtime)
        elif stat.S_ISDIR(s.st_mode):
            dirs.append(fullname)
    for fullname in dirs:
        for t in iter_files(fullname):
            yield t


class FileStore(object):
    """
    Arranges files in a special layout according to their content-hash.

    Improvements over original `FileStore`:

        1. All directories in layout are created when base directory is created
           meaning existence of directories doesn't need to be tested for later

        2. Because of (1), `FileStore.path()` and similar no longer require the
           ``create=True`` keyword argument

        3. `FileStore.__init__()` prevents you from creating a `FileStore`
           inside another `FileStore`

        4. Canonical names no longer include the file extension, which
           simplifies the code, eliminates a potential security risk, and allows
           us to provide even stronger data durability guarantees
    """

    def __init__(self, parent, dotname=DOTNAME):
        """
        Initialize.

        :param parent: directory for file store, eg ``'/home/jderose'``
        :param dotname: name of control directory, default is ``'.dmedia'``
        """
        self.parent = safe_path(parent)
        self.dotname = dotname
        self.base = path.join(self.parent, dotname)
        self.tmp = path.join(self.base, 'tmp')

        # Make sure parent isn't nested in another FileStore:
        if dotname in self.parent.split(os.sep):
            raise NestedParent(self.parent, dotname)

        # Make sure parent is a directory:
        if not path.isdir(self.parent):
            raise ValueError('{}.parent not a directory: {!r}'.format(
                self.__class__.__name__, self.parent)
            )

        # If base doesn't exist, create it and initialize all dirs in layout:
        if ensuredir(self.base):
            self.init_dirs()

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.parent)

    def __iter__(self):
        """
        Iterate through file IDs for all files in this filestore.

        This will yield the file ID in sorted order for all files in this
        filestore.
        """
        for prefix in B32NAMES:
            d = path.join(self.base, 'files', prefix)
            for name in sorted(os.listdir(d)):
                yield prefix + name

    def init_dirs(self):
        """
        Creates the needed subdirectories inside the control directory.

        `FileStore.__init__()` automatically calls this if the control directory
        does not yet exists.
        """
        for name in ('files', 'partial', 'corrupt', 'tmp'):
            d = path.join(self.base, name)
            ensuredir(d)
        for name in B32NAMES:
            d = path.join(self.base, 'files', name)
            ensuredir(d)
        f = path.join(self.base, 'README.txt')
        if not path.exists(f):
            open(f, 'wt').write(README)
            os.chmod(f, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)

    def catch_traversal(self, untrusted):
        """
        Verify that *untrusted* is inside this `FileStore` base directory.
        """
        abspath = path.abspath(untrusted)
        if abspath.startswith(self.base + os.sep):
            return abspath
        raise PathTraversal(self.base, untrusted, abspath)

    def join(self, *parts):
        """
        Safely join *parts* with base directory to prevent path traversal.

        For security reasons, it's very important that you use this method
        rather than ``path.join()`` directly.  This method will prevent path
        traversal attacks, ``path.join()`` will not.

        Also see `FileStore.catch_traversal()`.
        """
        return self.catch_traversal(path.join(self.base, *parts))

    def path(self, _id):
        """
        Get the canonical path of a file with *_id*.

        :param _id: base32-encoded content-hash
        """
        _id = safe_id(_id)
        return self.join('files', _id[:2], _id[2:])

    def partial_path(self, _id):
        """
        Get the canonical path of a partial file with *_id*.

        "Partial" files are files that are being transfered (either uploaded or
        downloaded).  Because transfers must be resume-able, they are given
        a consistent filename based on their *_id*.

        :param _id: base32-encoded content-hash
        """
        return self.join('partial', safe_id(_id))

    def corrupt_path(self, _id):
        """
        Get the canonical path of a corrupt file with *_id*.

        When `FileStore.verify()` finds a file to be corrupted, it moves the
        file into a special holding location based on the file's *_id*.

        The file isn't immediately deleted in case the intact leaves in the file
        are needed to reconstruct a fully valid file.  For example, if you have
        two copies of a file, each which is corrupted, you could still recreate
        the full file as long as the corruption wasn't in the same 8 MiB leaf.

        :param _id: base32-encoded content-hash
        """
        return self.join('corrupt', safe_id(_id))

    def exists(self, _id):
        """
        Return ``True`` if a file with *_id* exists.

        :param _id: base32-encoded content-hash
        """
        return path.isfile(self.path(_id))

    def open(self, _id):
        """
        Open the file with *_id* in ``'rb'`` mode.

        :param _id: base32-encoded content-hash
        """
        return open(self.path(_id), 'rb')

    def verify(self, _id, return_fp=False):
        """
        Verify integrity of file with ID *_id*.

        If the file's root-hash does not equal *_id*, a `FileIntegrityError`
        is raised.

        Otherwise a ContentHash namedtuple is returned with ``id``,
        ``file_size``, and ``leaf_hashes`` attributes.

        Or if called with ``return_fp=True``, the open file is returned after
        calling ``src_fp.seek(0)`` to put read position back at the start of
        the file.

        :param _id: base32-encoded content-hash
        :param return_fp: If ``True``, return open file, if ``False``, return
            a ContentHash namedtuple
        """
        src_fp = self.open(_id)
        c = hash_fp(src_fp)
        if c.id != _id:
            corrupt = self.move_to_corrupt(src_fp, _id)
            raise FileIntegrityError(self.parent, _id, c.id)
        if return_fp:
            src_fp.seek(0)
            return src_fp
        return c

    def verify_iter(self, _id, file_size, leaf_hashes):
        """
        Iterate through leaves as they are verified.

        This method will yield ``(leaf_index, leaf_data)`` tuples for each leaf
        in the file identified by *_id*.  This method verifies the ``leaf_data``
        prior to yielding it... so you can trust that the ``leaf_data`` is
        valid by the time you get it.

        The use case that inspired this method was calculating the file md5sum
        prior to uploading it to Amazon S3 using ``boto``.  If the file is
        corrupted, all the md5sum does is make sure the exact same corruption
        arrives intact on S3.

        :param _id: base32-encoded content-hash (an ``str``)
        :param file_size: file size in bytes (an ``int`` > 0)
        :param leaf_hashes: a ``bytes`` instance containing the concatenated
            leaf hashes
        """
        check_root_hash(_id, file_size, leaf_hashes)
        src_fp = self.open(_id)
        size = os.fstat(src_fp.fileno()).st_size
        if size != file_size:
            corrupt = self.move_to_corrupt(src_fp, _id)
            raise SizeIntegrityError(self.parent, _id, file_size, size)
        for (i, leaf) in reader_iter(src_fp):
            got = hash_leaf(i, leaf)
            expected = get_leaf_hash(leaf_hashes, i)
            if got != expected:
                self.move_to_corrupt(src_fp, _id)
                raise LeafIntegrityError(self.parent, _id, i, expected, got)
            yield (i, leaf)
        assert get_leaf_hash(leaf_hashes, i + 1) == b''

    def verify_iter2(self, _id):
        """
        Yield each leaf as it's read, verifying file integrity after final leaf.

        This method will yield ``(leaf_index, leaf_data)`` tuple for each leaf
        in the file identified by *_id*.  Use this method with care as the
        integrity of the leaves is not known till after the last leaf has been
        yielded.

        This method is similar to `FileStore.verify_iter()` except only the
        *_id* is needed.
        """
        src_fp = self.open(_id)
        h = Hasher()
        for (i, leaf) in reader_iter(src_fp):
            h.update(leaf)
            yield (i, leaf)
        c = h.content_hash()
        if c.id != _id:
            self.move_to_corrupt(src_fp, _id)
            raise FileIntegrityError(self.parent, _id, c.id)

    def remove(self, _id):
        """
        Delete file with *_id* from underlying filesystem.

        :param _id: base32-encoded content-hash
        """
        filename = self.path(_id)
        log.info('Removing %r from %r', _id, self)
        os.remove(filename)

    def allocate_tmp(self, size=None):
        """
        Allocate randomly named temporary file for an import or render.

        When importing a file, supply the file *size* in bytes, and that space
        will be efficiently pre-allocated.

        When rendering, the file size will not be known in advance, so *size*
        should not be supplied.

        :param size: file size in bytes (> 0)
        """
        (fileno, filename) = tempfile.mkstemp(dir=self.tmp)
        if size is not None:
            fallocate(size, filename)
        # FIXME: This probably isn't the best approach, but for now it works:
        tmp_fp = open(filename, 'r+b')
        os.close(fileno)
        return tmp_fp

    def allocate_partial(self, size, _id):
        """
        Allocate canonically named partial file for a transfer.

        :param size: file size in bytes (> 0)
        :param _id: base32-encoded content-hash
        """
        filename = self.partial_path(_id)
        fallocate(size, filename)
        try:
            fp = open(filename, 'r+b')
            if os.fstat(fp.fileno()).st_size > size:
                fp.truncate(size)
            return fp
        except IOError:
            return open(filename, 'wb')

    def move_to_canonical(self, src_fp, _id):
        """
        Move temporary file into its canonical location.

        This method will securely and atomically move a temporary file into its
        canonical location.

        Note, however, that this method does *not* verify the content hash of
        the temporary file!  This is by design as many operations will compute
        the content hash as they write to the temporary file.

        Regardless, the full content hash should have been verified prior to
        calling this method.  To ensure the content is not modified, operations
        must take these steps:

            1. Open *src_fp* and keep it open, thereby retaining a lock on the
               file

            2. Compute the full content hash, which can be done as content is
               written to *src_fp* (open in mode ``'r+b'`` to resume a transfer,
               but hash of previously transfered leaves must still be verified)

            3. With *src_fp* still open, move the temporary file into its
               canonical location using this method.

        As a simple locking mechanism, this method takes an open ``file`` rather
        than a filename, thereby preventing the file from being modified during
        the move.  A ``ValueError`` is raised if *src_fp* is already closed.

        For portability reasons, this method requires that *src_fp* be opened in
        a binary mode: ``'rb'``, ``'wb'``, or ``'rb+'``.  A ``ValueError`` is
        raised if opened in any other mode.

        For security reasons, this method will only move a temporary file
        located within the ``FileStore.base`` directory or a subdirectory
        thereof.  If an attempt is made to move a file from outside the store,
        `PathTraversal` is raised.  See `FileStore.catch_traversal()`.

        Just prior to moving the file, a call to ``os.fchmod()`` is made to set
        read-only permissions (0444).  After the move, *src_fp* is closed.

        The return value is the absolute path of the canonical file.

        :param src_fp: a file opened in ``'wb'``, ``'rb'``, or ``'rb+'`` mode
        :param _id: base32-encoded dmedia content-hash
        """
        if not isinstance(src_fp, io.BufferedIOBase):
            raise TypeError(TYPE_ERROR.format(
                    'src_fp', io.BufferedIOBase, type(src_fp), src_fp
                )
            )
        assert src_fp.mode in ('wb', 'rb', 'rb+')
        if src_fp.closed:
            raise ValueError(
                'src_fp is closed, must be open: {!r}'.format(src_fp)
            )
        src = self.catch_traversal(src_fp.name)
        dst = self.path(_id)
        # FIXME: Still need to decide the *new* way of handling duplicates in a
        # FileStore... although the *new* way might be the current way below
        src_fp.flush()
        fileno = src_fp.fileno()
        os.fsync(fileno)  # FIXME: big IO performance hit because of this?
        os.fchmod(fileno, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
        log.info('Moving file %r to %r', src, dst)
        os.rename(src, dst)
        src_fp.close()
        return dst

    def move_to_corrupt(self, src_fp, _id):
        """
        Move a file from its canonical location to its corrupt location.

        While a file is found to be corrupt (meaning the computed content hash
        doesn't match the expected content hash), it is moved from its canonical
        location to a special corrupt location.

        For data safety and forensics, corrupt files are not immediately
        deleted.  As files can be verified with 8 MiB granularity, it is
        possible to recover the complete file from two separate corrupt copies
        as long as the corruption is in different 8 MiB leaves.

        This method returns the full path of the corrupt file.

        :param src_fp: a file opened in ``'rb'`` mode
        :param _id: base32-encoded dmedia content-hash
        """
        if not isinstance(src_fp, io.BufferedReader):
            raise TypeError(TYPE_ERROR.format(
                    'src_fp', io.BufferedReader, type(src_fp), src_fp
                )
            )
        assert src_fp.mode == 'rb'
        if src_fp.closed:
            raise ValueError(
                'src_fp is closed, must be open: {!r}'.format(src_fp)
            )
        src = self.catch_traversal(src_fp.name)
        dst = self.corrupt_path(_id)
        log.info('Moving file %r to %r', src, dst)
        os.rename(src, dst)
        src_fp.close()
        return dst

    def import_file(self, src_fp):
        """
        Atomically copy open file *src_fp* into this filestore.

        The method will compute the content-hash of *src_fp* as it copies it to
        a temporary file within this store.  Once the copying is complete, the
        temporary file will be moved to its canonical location using
        `FileStore.move_to_canonical()`.

        This method returns an ``(id, file_size, leaf_hashes)`` namedtuple.

        Note that *src_fp* must have been opened in ``'rb'`` mode.

        :param src_fp: a ``file`` instance created with ``open()``
        """
        size = os.fstat(src_fp.fileno()).st_size
        tmp_fp = self.allocate_tmp(size)
        log.info('Importing file %r into %r', src_fp.name, self)
        c = hash_fp(src_fp, tmp_fp)
        assert c.file_size == size
        self.move_to_canonical(tmp_fp, c.id)
        return c

    def copy(self, _id, *filestores):
        """
        Copy file with *_id* from this FileStore into one or more *filestores*.
        """
        if not filestores:
            raise TypeError(
                '{}.copy() takes at least one destination filestore'.format(
                    self.__class__.__name__
                )
            )
        for (i, fs) in enumerate(filestores):
            if not isinstance(fs, FileStore):
                name = 'filestores[{}]'.format(i)
                raise TypeError(
                    TYPE_ERROR.format(name, FileStore, type(fs), fs)
                )
            if fs.parent == self.parent:
                name = 'filestores[{}]'.format(i)
                raise ValueError(
                    'wont copy into {}, same parent as {!r}'.format(name, self)
                )
        src_fp = self.open(_id)
        size = os.fstat(src_fp.fileno()).st_size
        temps = [fs.allocate_tmp(size) for fs in filestores]
        h = Hasher()
        for (i, leaf) in reader_iter(src_fp):
            h.update(leaf)
            for tmp_fp in temps:
                tmp_fp.write(leaf)
        c = h.content_hash()
        if c.id != _id:
            self.move_to_corrupt(src_fp, _id)
            for tmp_fp in temps:
                tmp_fp.close()
                os.remove(tmp_fp.name)
            raise FileIntegrityError(self.parent, _id, c.id)
        assert c.file_size == size
        for (i, fs) in enumerate(filestores):
            tmp_fp = temps[i]
            fs.move_to_canonical(tmp_fp, _id)
        return c

    def copy_all(self, *filestores):
        """
        Copy all files in this FileStore into one or more other *filestores*.
        """
        for _id in self:
            try:
                yield self.copy(_id, *filestores)
            except IntegrityError as e:
                yield e
