# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2011 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for `filestore` module.
"""

from unittest import TestCase
import os
from os import path
import io
import stat
from base64 import b32encode
from subprocess import check_call
import tempfile
import shutil

from skein import skein512

import filestore


TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'

B32NAMES = []
for a in filestore.B32ALPHABET:
    for b in filestore.B32ALPHABET:
        B32NAMES.append(a + b)
B32NAMES.sort()
B32NAMES = tuple(B32NAMES)


# Sample file:
LEAVES = (
    b'a' * filestore.LEAF_SIZE,
    b'b' * filestore.LEAF_SIZE,
    b'c' * (filestore.LEAF_SIZE // 2),
)
FILE_SIZE = (filestore.LEAF_SIZE * 2) + (filestore.LEAF_SIZE // 2)
LEAF_HASHES = b''.join([
    filestore.hash_leaf(0, LEAVES[0]),
    filestore.hash_leaf(1, LEAVES[1]),
    filestore.hash_leaf(2, LEAVES[2]),
])
ID = filestore.hash_root(FILE_SIZE, LEAF_HASHES)
CH = filestore.ContentHash(ID, FILE_SIZE, LEAF_HASHES)
assert CH.id == 'YDDL5ROVABZP4NBSJPC3HUQDVDAGAP5L26YFXD3UR6N5OLVN'


def write_sample_file(dst):
    fp = open(dst, 'wb')
    for leaf in LEAVES:
        fp.write(leaf)
    fp.close()


def tobytes(n):
    return str(n).encode('utf-8')


def random_id():
    return b32encode(os.urandom(filestore.DIGEST_BYTES)).decode('ascii')


class TempDir(object):
    def __init__(self, prefix='unit-tests.'):
        self.dir = tempfile.mkdtemp(prefix=prefix)

    def join(self, *parts):
        return path.join(self.dir, *parts)

    def rmtree(self):
        if self.dir is not None:
            check_call(['/bin/chmod', '-R', '+w', self.dir])
            shutil.rmtree(self.dir)
            self.dir = None

    def makedirs(self, *parts):
        d = self.join(*parts)
        if not path.exists(d):
            os.makedirs(d)
        return d

    def touch(self, *parts):
        d = self.makedirs(*parts[:-1])
        f = self.join(*parts)
        assert not path.exists(f)
        open(f, 'wb').close()
        assert path.isfile(f) and not path.islink(f)
        return f

    def write(self, content, *parts):
        d = self.makedirs(*parts[:-1])
        f = self.join(*parts)
        assert not path.exists(f)
        open(f, 'wb').write(content)
        assert path.isfile(f) and not path.islink(f)
        return f

    def copy(self, src, *parts):
        self.makedirs(*parts[:-1])
        dst = self.join(*parts)
        assert not path.exists(dst)
        shutil.copy2(src, dst)
        assert path.isfile(dst) and not path.islink(dst)
        return dst

    def __del__(self):
        self.rmtree()


class TestRootHashError(TestCase):
    def test_init(self):
        e = filestore.RootHashError('foo', 18, b'leaf hashes', 'bar')
        self.assertEqual(e.id, 'foo')
        self.assertEqual(e.file_size, 18)
        self.assertEqual(e.leaf_hashes, b'leaf hashes')
        self.assertEqual(e.bad_id, 'bar')
        self.assertEqual(str(e), "'foo'")


class TestNestedParent(TestCase):
    def test_init(self):
        e = filestore.NestedParent('/home/joe/.dmedia/sue', '.dmedia')
        self.assertEqual(e.parent, '/home/joe/.dmedia/sue')
        self.assertEqual(e.dotname, '.dmedia')
        self.assertEqual(str(e), "'/home/joe/.dmedia/sue' contains '.dmedia'")


class TestAmbigousPath(TestCase):
    def test_init(self):
        e = filestore.AmbiguousPath('/foo/../bar', '/bar')
        self.assertEqual(e.pathname, '/foo/../bar')
        self.assertEqual(e.abspath, '/bar')
        self.assertEqual(str(e), "'/foo/../bar' resolves to '/bar'")


class TestPathTraversal(TestCase):
    def test_init(self):
        e = filestore.PathTraversal(
            '/home/user/.dmedia',
            '/home/user/.dmeida/files/../../.ssh',
            '/home/user/.ssh'
        )
        self.assertEqual(e.base, '/home/user/.dmedia')
        self.assertEqual(e.pathname, '/home/user/.dmeida/files/../../.ssh')
        self.assertEqual(e.abspath, '/home/user/.ssh')
        self.assertEqual(
            str(e),
            "'/home/user/.ssh' outside of '/home/user/.dmedia'"
        )


class TestIntegrityError(TestCase):
    def test_init(self):
        e = filestore.IntegrityError('/the/parent', 'foo')
        self.assertEqual(e.parent, '/the/parent')
        self.assertEqual(e.id, 'foo')
        self.assertEqual(
            str(e),
            "'foo' is corrupted in '/the/parent'"
        )


class TestFileIntegrityError(TestCase):
    def test_init(self):
        e = filestore.FileIntegrityError('/the/parent', 'foo', 'bar')
        self.assertEqual(e.parent, '/the/parent')
        self.assertEqual(e.id, 'foo')
        self.assertEqual(e.bad_id, 'bar')
        self.assertEqual(
            str(e),
            "'foo' is corrupted in '/the/parent'"
        )


class TestLeafIntegrityError(TestCase):
    def test_init(self):
        e = filestore.LeafIntegrityError('/the/parent', 'foo', 3, 'good', 'bad')
        self.assertEqual(e.parent, '/the/parent')
        self.assertEqual(e.id, 'foo')
        self.assertEqual(e.leaf_index, 3)
        self.assertEqual(e.leaf_hash, 'good')
        self.assertEqual(e.bad_leaf_hash, 'bad')
        self.assertEqual(
            str(e),
            "'foo' is corrupted in '/the/parent'"
        )


class TestSizeIntegrityError(TestCase):
    def test_init(self):
        e = filestore.SizeIntegrityError('/the/parent', 'foo', 234, 345)
        self.assertEqual(e.parent, '/the/parent')
        self.assertEqual(e.id, 'foo')
        self.assertEqual(e.file_size, 234)
        self.assertEqual(e.bad_file_size, 345)
        self.assertEqual(
            str(e),
            "'foo' is corrupted in '/the/parent'"
        )


class TestFunctions(TestCase):
    def test_constants(self):
        """
        Sanity checks on various constants.
        """
        self.assertEqual(
            filestore.B32ALPHABET,
            ''.join(sorted(filestore.B32ALPHABET))
        )

        self.assertEqual(filestore.B32NAMES, B32NAMES)
        self.assertEqual(len(filestore.B32NAMES), 1024)

        self.assertTrue(filestore.DOTNAME.startswith('.'))

        # Must be multiple of 40 bits for base32-encoding
        self.assertEqual(filestore.DIGEST_BITS % 40, 0)

        # And a multiple of 8 bits, of course
        self.assertEqual(filestore.DIGEST_BITS % 8, 0)

        self.assertEqual(filestore.DIGEST_BYTES * 8, filestore.DIGEST_BITS)

        # Must be multiple of 1 MiB
        self.assertEqual(filestore.LEAF_SIZE % (2**20), 0)


    def test_hash_leaf(self):
        f = filestore.hash_leaf

        # Test with wrong leaf_index type:
        with self.assertRaises(TypeError) as cm:
            pers = f(0.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 0.0)
        )
        with self.assertRaises(TypeError) as cm:
            pers = f(1.5)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_index', int, float, 1.5)
        )

        # Test with wrong leaf_index value
        with self.assertRaises(ValueError) as cm:
            pers = f(-1)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be >= 0; got -1'
        )
        with self.assertRaises(ValueError) as cm:
            pers = f(-17)
        self.assertEqual(
            str(cm.exception),
            'leaf_index: must be >= 0; got -17'
        )

        # Test with wrong leaf_data type
        leaf_data = 'a' * 31415
        with self.assertRaises(TypeError) as cm:
            digest = f(2, leaf_data)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_data', bytes, str, leaf_data)
        )

        # Test with wrong leaf_data length
        toosmall = b''
        with self.assertRaises(ValueError) as cm:
            digest = f(2, toosmall)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= len(leaf_data) <= 8388608; got 0'
        )
        toobig = b'a' * 8388609
        with self.assertRaises(ValueError) as cm:
            digest = f(2, toobig)
        self.assertEqual(
            str(cm.exception),
            'Need 1 <= len(leaf_data) <= 8388608; got 8388609'
        )

        # Test with good values:
        lh = f(2)
        self.assertEqual(lh.block_bits, 512)
        self.assertEqual(lh.digest_bits, 240)
        self.assertEqual(
            lh.digest(),
            skein512(b'2',
                digest_bits=240,
                pers=filestore.PERS_LEAF,
            ).digest()
        )

        content = b'N'
        digest = f(2, content)
        self.assertEqual(
            digest,
            skein512(b'2' + content,
                digest_bits=240,
                pers=filestore.PERS_LEAF,
            ).digest()
        )

        content = b'N' * filestore.LEAF_SIZE
        digest = f(2, content)
        self.assertEqual(
            digest,
            skein512(b'2' + content,
                digest_bits=240,
                pers=filestore.PERS_LEAF,
            ).digest()
        )

        # A 2k value sanity check on our crytographic claim that the leaf_index
        # is tied to the leaf_hash:
        leaf = b'N' * 314159
        leaf_hash = filestore.hash_leaf(1000, leaf)
        for i in range(1, 1001):
            self.assertNotEqual(filestore.hash_leaf(1000 - i, leaf), leaf_hash)
            self.assertNotEqual(filestore.hash_leaf(1000 + i, leaf), leaf_hash)

        # A 2k random value sanity check on our crytographic claim that the leaf
        # is tied to the leaf_hash:
        for i in range(2000):
            l = leaf[:-15] + os.urandom(15)
            self.assertEqual(len(l), len(leaf))
            self.assertNotEqual(filestore.hash_leaf(1000, l), leaf_hash)

    def test_hash_root(self):
        f = filestore.hash_root

        # Test with wrong file_size type:
        with self.assertRaises(TypeError) as cm:
            pers = f(1.0, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.0)
        )
        with self.assertRaises(TypeError) as cm:
            pers = f(1.5, None)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('file_size', int, float, 1.5)
        )

        # Test with wrong file_size value
        with self.assertRaises(ValueError) as cm:
            pers = f(0, None)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be >= 1; got 0'
        )
        with self.assertRaises(ValueError) as cm:
            pers = f(-1, None)
        self.assertEqual(
            str(cm.exception),
            'file_size: must be >= 1; got -1'
        )

        # Test with good file_size and leaf_hashes
        self.assertEqual(
            f(14, filestore.hash_leaf(0, b'Hello Novacut!')),
            'CBQR3PKOTFLKCF4DWTJVVC4WZSTZV3SSWD65VCLJWTKMPVGO'
        )

        # A 100k value sanity check on our crytographic claim that the
        # file_size is tied to the root_hash:
        root_hash = filestore.hash_root(FILE_SIZE, LEAF_HASHES)
        for i in range(1, 50001):
            self.assertNotEqual(
                filestore.hash_root(FILE_SIZE - i, LEAF_HASHES),
                root_hash
            )
            self.assertNotEqual(
                filestore.hash_root(FILE_SIZE + i, LEAF_HASHES),
                root_hash
            )

        # A 10k random value sanity check on our crytographic claim that the
        # leaf_hashes are tied to the root_hash:
        for i in range(10000):
            leaf_hashes = os.urandom(len(LEAF_HASHES))
            self.assertNotEqual(
                filestore.hash_root(FILE_SIZE, leaf_hashes),
                root_hash
            )

    def test_check_leaf_hashes(self):
        f = filestore.check_leaf_hashes

        # Test with wrong type:
        leaf_hashes = 'a' * 60
        with self.assertRaises(TypeError) as cm:
            l = f(leaf_hashes)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('leaf_hashes', bytes, str, leaf_hashes)
        )

        # Test with empty leaf_hashes:
        with self.assertRaises(ValueError) as cm:
            l = f(b'')
        self.assertEqual(str(cm.exception), 'leaf_hashes cannot be empty')

        # Test with wrong length:
        leaf_hashes = b'a' * 61
        with self.assertRaises(ValueError) as cm:
            l = f(leaf_hashes)
        self.assertEqual(
            str(cm.exception),
            'len(leaf_hashes) must be multiple of 30; got 61: {!r}'.format(leaf_hashes)
        )

        # Test with good values
        good = b'a' * 60
        self.assertIs(f(good), good)

        array = bytearray()
        for x in range(256):
            array.extend(x for i in range(30))
        big = bytes(array)
        self.assertEqual(len(big), 30 * 256)
        self.assertIs(f(big), big)

    def test_check_root_hash(self):
        f = filestore.check_root_hash

        self.assertEqual(f(ID, FILE_SIZE, LEAF_HASHES), CH)

        with self.assertRaises(filestore.RootHashError) as cm:
            f(ID, FILE_SIZE + 1, LEAF_HASHES)
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.file_size, FILE_SIZE + 1)
        self.assertEqual(cm.exception.leaf_hashes, LEAF_HASHES)
        self.assertEqual(cm.exception.bad_id,
            '7XSYWHJVRWGEMWN5LNFSIUVXVBURNMKF4XDRXR2T27DZMHPP'
        )

        with self.assertRaises(filestore.RootHashError) as cm:
            f(ID, FILE_SIZE - 1, LEAF_HASHES)
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.file_size, FILE_SIZE - 1)
        self.assertEqual(cm.exception.leaf_hashes, LEAF_HASHES)
        self.assertEqual(cm.exception.bad_id,
            'CMVK6EFJMA32ZI4JRXED564PNRJKXBT7MFA436IA7NOWR6CM'
        )

    def test_enumerate_leaf_hashes(self):
        f = filestore.enumerate_leaf_hashes
        self.assertEqual(
            list(f(b'a' * 30)),
            [(0, b'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')]
        )
        leaf_hashes = (b'X' * 30) + (b'Y' * 30) + (b'Z' * 30)
        self.assertEqual(
            list(f(leaf_hashes)),
            [
                (0, b'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'),
                (1, b'YYYYYYYYYYYYYYYYYYYYYYYYYYYYYY'),
                (2, b'ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ'),
            ]
        )

    def test_get_leaf_hash(self):
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 0),
            filestore.hash_leaf(0, LEAVES[0])
        )
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 1),
            filestore.hash_leaf(1, LEAVES[1])
        )
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 2),
            filestore.hash_leaf(2, LEAVES[2])
        )
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 3),
            b''
        )
        self.assertEqual(
            filestore.get_leaf_hash(LEAF_HASHES, 4),
            b''
        )

    def test_missing_leaves(self):
        f = filestore.missing_leaves
        tmp = TempDir()
        chunks = (
            b'Y' * filestore.LEAF_SIZE,
            b'E' * filestore.LEAF_SIZE,
            b'S',
        )
        leaf_hashes = b''.join(
            filestore.hash_leaf(i, chunk) for (i, chunk) in enumerate(chunks)
        )
        src = tmp.touch('test.mov')

        # Test with empty file
        fp = open(src, 'rb')
        self.assertEqual(
            list(f(fp, leaf_hashes)),
            list(filestore.enumerate_leaf_hashes(leaf_hashes))
        )

        # Test with correct file
        fp = open(src, 'wb')
        for chunk in chunks:
            fp.write(chunk)
        fp = open(src, 'rb')
        self.assertEqual(
            list(f(fp, leaf_hashes)),
            [],
        )

        # Test with just the last 1-byte leaf correct:
        fp = open(src, 'wb')
        fp.write(b'N' * filestore.LEAF_SIZE)
        fp.write(b'O' * filestore.LEAF_SIZE)
        fp.write(b'S')
        fp = open(src, 'rb')
        self.assertEqual(
            list(f(fp, leaf_hashes)),
            [
                (0, filestore.hash_leaf(0, chunks[0])),
                (1, filestore.hash_leaf(1, chunks[1])),
            ]
        )

    def test_safe_path(self):
        f = filestore.safe_path

        # Test with relative path:
        with self.assertRaises(filestore.AmbiguousPath) as cm:
            f('foo/bar')
        self.assertEqual(cm.exception.pathname, 'foo/bar')
        self.assertEqual(cm.exception.abspath, path.abspath('foo/bar'))

        # Test with path traversal:
        with self.assertRaises(filestore.AmbiguousPath) as cm:
            f('/foo/bar/../../root')
        self.assertEqual(cm.exception.pathname, '/foo/bar/../../root')
        self.assertEqual(cm.exception.abspath, '/root')

        # Test with normalized absolute path:
        self.assertEqual(f('/home/jderose/.dmedia'), '/home/jderose/.dmedia')

    def test_safe_id(self):
        # Test with wrong type
        with self.assertRaises(TypeError) as cm:
            filestore.safe_id(42)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('_id', str, int, 42)
        )

        # Test with invalid base32 encoding:
        bad = 'NWBNVXVK5DQGIOW7MYR4K3KA5K22W7N'
        with self.assertRaises(ValueError) as cm:
            filestore.safe_id(bad)
        self.assertEqual(
            str(cm.exception),
            '_id: cannot b32decode %r: Incorrect padding' % bad
        )

        # Test with wrong length:
        bad = 'NWBNVXVK5DQGIOW7MYR4K3KA'
        with self.assertRaises(ValueError) as cm:
            filestore.safe_id(bad)
        self.assertEqual(
            str(cm.exception),
            '_id: decoded must be 30 bytes; got 15: {!r}'.format(bad)
        )

        # Test with a good chash:
        good = 'DHS36R6HGG52TQRB6BN7N66EWOPU4H67WISD3J6TWNCXXAZT'
        self.assertIs(filestore.safe_id(good), good)

    def test_reader_iter(self):
        tmp = TempDir()
        src = tmp.write(b'hello', 'some', 'file.txt')
        allowed = (io.BufferedReader, io.BufferedRandom)

        # Wrong src_fp type:
        with self.assertRaises(TypeError) as cm:
            for (i, leaf) in filestore.reader_iter(17):
                pass
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', allowed, int, 17)
        )

        # File opened in wrong mode:
        src_fp = open(src, 'r')
        with self.assertRaises(TypeError) as cm:
            for (i, leaf) in filestore.reader_iter(src_fp):
                pass
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', allowed, type(src_fp), src_fp)
        )

    def test_hash_fp(self):
        tmp = TempDir()
        src = tmp.join('foo.mov')
        dst = tmp.join('bar.mov')
        write_sample_file(src)

        src_fp = open(src, 'rb')
        ch = filestore.hash_fp(src_fp)
        self.assertIsInstance(ch, filestore.ContentHash)
        self.assertEqual(ch.id,
            'YDDL5ROVABZP4NBSJPC3HUQDVDAGAP5L26YFXD3UR6N5OLVN'
        )
        self.assertEqual(ch.file_size, 20971520)
        self.assertFalse(path.exists(dst))

        src_fp = open(src, 'rb')
        dst_fp = open(dst, 'wb')
        ch = filestore.hash_fp(src_fp, dst_fp)
        self.assertIsInstance(ch, filestore.ContentHash)
        self.assertEqual(ch.id,
            'YDDL5ROVABZP4NBSJPC3HUQDVDAGAP5L26YFXD3UR6N5OLVN'
        )
        self.assertEqual(ch.file_size, 20971520)
        self.assertTrue(path.isfile(dst))

        src_fp = open(dst, 'rb')
        self.assertEqual(filestore.hash_fp(src_fp), ch)

    def test_fallocate(self):
        f = filestore.fallocate
        tmp = TempDir()
        filename = tmp.join('example.mov')

        # Test when size is wrong type:
        with self.assertRaises(TypeError) as cm:
            f('2311', filename)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('size', int, str, '2311')
        )

        # Test when size <= 0
        with self.assertRaises(ValueError) as cm:
            f(0, filename)
        self.assertEqual(str(cm.exception), 'size must be >0; got 0')
        with self.assertRaises(ValueError) as cm:
            f(-2311, filename)
        self.assertEqual(str(cm.exception), 'size must be >0; got -2311')

        # Test with relative path:
        with self.assertRaises(filestore.AmbiguousPath) as cm:
            f(2311, 'foo/bar')
        self.assertEqual(cm.exception.pathname, 'foo/bar')
        self.assertEqual(cm.exception.abspath, path.abspath('foo/bar'))

        # Test with path traversal:
        with self.assertRaises(filestore.AmbiguousPath) as cm:
            f(2311, '/foo/bar/../../root')
        self.assertEqual(cm.exception.pathname, '/foo/bar/../../root')
        self.assertEqual(cm.exception.abspath, '/root')

        # Test with correct args:
        self.assertFalse(path.exists(filename))
        ret = f(2311, filename)
        self.assertTrue(ret in [None, True, False])

        if ret is None:
            self.assertFalse(path.exists(filename))

        if ret is True:
            self.assertTrue(path.exists(filename))
            self.assertEqual(path.getsize(filename), 2311)

        if ret is False:
            self.assertTrue(path.exists(filename))
            self.assertEqual(path.getsize(filename), 0)

    def test_ensuredir(self):
        f = filestore.ensuredir
        tmp = TempDir()

        # Test that os.makedirs() is not used:
        d = tmp.join('foo', 'bar')
        with self.assertRaises(ValueError) as cm:
            f(d)
        self.assertEqual(str(cm.exception), 'not a directory: {!r}'.format(d))
        self.assertFalse(path.exists(d))
        self.assertFalse(path.exists(tmp.join('foo')))

        # Test when dir does not exist
        d = tmp.join('foo')
        self.assertFalse(path.exists(d))
        self.assertIs(f(d), True)
        self.assertTrue(path.isdir(d))

        # Test when dir exists:
        self.assertIs(f(d), False)
        self.assertTrue(path.isdir(d))

        # Test when dir exists and is a file:
        d = tmp.touch('bar')
        with self.assertRaises(ValueError) as cm:
            f(d)
        self.assertEqual(str(cm.exception), 'not a directory: {!r}'.format(d))

        # Test when dir exists and is a symlink to a dir:
        target = tmp.makedirs('target')
        link = tmp.join('link')
        os.symlink(target, link)
        self.assertTrue(path.isdir(link))
        self.assertTrue(path.islink(link))
        with self.assertRaises(ValueError) as cm:
            f(link)
        self.assertEqual(
            str(cm.exception),
            '{!r} is symlink to {!r}'.format(link, target)
        )

    def test_iter_files(self):
        tmp = TempDir()
        self.assertEqual(list(filestore.iter_files(tmp.dir)), [])

        good = tmp.makedirs('good')
        bad = tmp.touch('bad')
        nope = tmp.join('nope')

        # Test when base dose not exist:
        with self.assertRaises(OSError) as cm:
            list(filestore.iter_files(nope))
        self.assertEqual(cm.exception.errno, 2)

        # Test when base is a file rather than directory:
        with self.assertRaises(OSError) as cm:
            list(filestore.iter_files(bad))
        self.assertEqual(cm.exception.errno, 20)

        self.assertEqual(list(filestore.iter_files(good)), [])

        b = tmp.makedirs('good', 'b')
        c = tmp.write(b'n' * 18, 'good', 'c')
        a = tmp.write(b'n' * 314, 'good', 'a')
        d = tmp.write(b'n' * 42, 'good', 'd')

        b_a = tmp.write(b'n' * 32, 'good', 'b', 'a')
        b_b = tmp.write(b'n' * 64, 'good', 'b', 'b')
        b_c = tmp.write(b'n' * 8, 'good', 'b', 'c')
        b_d = tmp.write(b'n' * 16, 'good', 'b', 'd')

        # All symlinks should be ignored:
        bb = tmp.join('good', 'bb')
        os.symlink(b, bb)
        self.assertTrue(path.isdir(bb))
        self.assertEqual(os.readlink(bb), b)

        os.symlink(a, tmp.join('good', 'aa'))
        os.symlink(c, tmp.join('good', 'cc'))
        os.symlink(d, tmp.join('good', 'dd'))

        self.assertEqual(
            list(filestore.iter_files(good)),
            [
                filestore.File(a, 314, path.getmtime(a)),
                filestore.File(c, 18, path.getmtime(c)),
                filestore.File(d, 42, path.getmtime(d)),

                filestore.File(b_a, 32, path.getmtime(b_a)),
                filestore.File(b_b, 64, path.getmtime(b_b)),
                filestore.File(b_c, 8, path.getmtime(b_c)),
                filestore.File(b_d, 16, path.getmtime(b_d)),
            ]
        )


class TestHasher(TestCase):
    def test_init(self):
        h = filestore.Hasher()
        self.assertEqual(h.file_size, 0)
        self.assertEqual(h.leaf_index, 0)
        self.assertIsInstance(h.array, bytearray)
        self.assertEqual(h.array, b'')
        self.assertFalse(h.closed)

    def test_update(self):
        h = filestore.Hasher()

        self.assertEqual(h.update(LEAVES[0]), LEAF_HASHES[0:30])
        self.assertEqual(h.leaf_index, 1)
        self.assertEqual(h.file_size, filestore.LEAF_SIZE)
        self.assertFalse(h.closed)

        self.assertEqual(h.update(LEAVES[1]), LEAF_HASHES[30:60])
        self.assertEqual(h.leaf_index, 2)
        self.assertEqual(h.file_size, filestore.LEAF_SIZE * 2)
        self.assertFalse(h.closed)

        self.assertEqual(h.update(LEAVES[2]), LEAF_HASHES[60:90])
        self.assertEqual(h.leaf_index, 3)
        self.assertEqual(h.file_size, sum(len(l) for l in LEAVES))
        self.assertTrue(h.closed)

    def test_content_hash(self):
        h = filestore.Hasher()
        for l in LEAVES:
            h.update(l)
        self.assertEqual(h.content_hash(), CH)


class TestFileStore(TestCase):
    def test_init(self):

        # Test with relative path:
        with self.assertRaises(filestore.AmbiguousPath) as cm:
            fs = filestore.FileStore('foo/bar')
        self.assertEqual(cm.exception.pathname, 'foo/bar')
        self.assertEqual(cm.exception.abspath, path.abspath('foo/bar'))
        self.assertEqual(
            str(cm.exception),
            '{!r} resolves to {!r}'.format('foo/bar', path.abspath('foo/bar'))
        )

        # Test with path traversal:
        with self.assertRaises(filestore.AmbiguousPath) as cm:
            fs = filestore.FileStore('/foo/bar/../../root')
        self.assertEqual(cm.exception.pathname, '/foo/bar/../../root')
        self.assertEqual(cm.exception.abspath, '/root')
        self.assertEqual(
            str(cm.exception),
            '{!r} resolves to {!r}'.format('/foo/bar/../../root', '/root')
        )

        # Test with a nested parent and default dotname:
        tmp = TempDir()
        parent = tmp.makedirs('.dmedia', 'foo')
        with self.assertRaises(filestore.NestedParent) as cm:
            fs = filestore.FileStore(parent)
        self.assertEqual(cm.exception.parent, parent)
        self.assertEqual(cm.exception.dotname, '.dmedia')
        self.assertEqual(
            str(cm.exception),
            '{!r} contains {!r}'.format(parent, '.dmedia')
        )

        # Test with a nested parent and custom dotname:
        tmp = TempDir()
        parent = tmp.makedirs('.baz', 'bar')
        with self.assertRaises(filestore.NestedParent) as cm:
            fs = filestore.FileStore(parent, '.baz')
        self.assertEqual(cm.exception.parent, parent)
        self.assertEqual(cm.exception.dotname, '.baz')
        self.assertEqual(
            str(cm.exception),
            '{!r} contains {!r}'.format(parent, '.baz')
        )

        # Test when parent doesn't exists:
        tmp = TempDir()
        parent = tmp.join('nope')
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(parent)
        self.assertEqual(
            str(cm.exception),
            'FileStore.parent not a directory: {!r}'.format(parent)
        )

        # Test when parent is a file:
        tmp = TempDir()
        parent = tmp.touch('somefile')
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(parent)
        self.assertEqual(
            str(cm.exception),
            'FileStore.parent not a directory: {!r}'.format(parent)
        )

        # Test when base does not exist
        tmp = TempDir()
        base = tmp.join('.dmedia')
        self.assertFalse(path.exists(base))
        fs = filestore.FileStore(tmp.dir)
        self.assertEqual(fs.parent, tmp.dir)
        self.assertEqual(fs.dotname, '.dmedia')
        self.assertEqual(fs.base, base)
        self.assertTrue(path.isdir(base))
        self.assertEqual(fs.tmp, path.join(base, 'tmp'))
        self.assertTrue(path.isdir(fs.tmp))

        # Test when base does not exist and custom dotname is used
        tmp = TempDir()
        base = tmp.join('.foo')
        self.assertFalse(path.exists(base))
        fs = filestore.FileStore(tmp.dir, '.foo')
        self.assertEqual(fs.parent, tmp.dir)
        self.assertEqual(fs.dotname, '.foo')
        self.assertEqual(fs.base, base)
        self.assertTrue(path.isdir(base))
        self.assertEqual(fs.tmp, path.join(base, 'tmp'))
        self.assertTrue(path.isdir(fs.tmp))

        # Test when base exists and is a directory
        tmp = TempDir()
        base = tmp.makedirs('.dmedia')
        fs = filestore.FileStore(tmp.dir)
        self.assertEqual(fs.parent, tmp.dir)
        self.assertEqual(fs.dotname, '.dmedia')
        self.assertEqual(fs.base, base)
        self.assertTrue(path.isdir(base))
        self.assertEqual(fs.tmp, path.join(base, 'tmp'))
        self.assertFalse(path.isdir(fs.tmp))

        # Test when base exists and is a directory, with custom dotname
        tmp = TempDir()
        base = tmp.makedirs('.bzr')
        fs = filestore.FileStore(tmp.dir, '.bzr')
        self.assertEqual(fs.parent, tmp.dir)
        self.assertEqual(fs.dotname, '.bzr')
        self.assertEqual(fs.base, base)
        self.assertTrue(path.isdir(base))
        self.assertEqual(fs.tmp, path.join(base, 'tmp'))
        self.assertFalse(path.isdir(fs.tmp))

        # Test when base exists and is a file
        tmp = TempDir()
        base = tmp.touch('.dmedia')
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(tmp.dir)
        self.assertEqual(
            str(cm.exception),
            'not a directory: {!r}'.format(base)
        )

        # Test when base exists and is a file, with custom dotname
        tmp = TempDir()
        base = tmp.touch('.hg')
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(tmp.dir, '.hg')
        self.assertEqual(
            str(cm.exception),
            'not a directory: {!r}'.format(base)
        )

        # Test when base exists and is a symlink to a dir
        tmp = TempDir()
        d = tmp.makedirs('foo')
        base = tmp.join('.dmedia')
        os.symlink(d, base)
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(tmp.dir)
        self.assertEqual(
            str(cm.exception),
            '{!r} is symlink to {!r}'.format(base, d)
        )

        # Test when base exists and is a symlink to a dir, with custom dotname
        tmp = TempDir()
        d = tmp.makedirs('bar')
        base = tmp.join('.git')
        os.symlink(d, base)
        with self.assertRaises(ValueError) as cm:
            fs = filestore.FileStore(tmp.dir, '.git')
        self.assertEqual(
            str(cm.exception),
            '{!r} is symlink to {!r}'.format(base, d)
        )

    def test_repr(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        self.assertEqual(repr(fs), 'FileStore({!r})'.format(tmp.dir))

    def test_iter(self):
        ids = tuple(random_id() for i in range(1000))
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        self.assertEqual(list(fs), [])
        for _id in ids:
            canonical = fs.path(_id)
            open(canonical, 'wb').write(_id.encode('utf-8'))
        self.assertEqual(list(fs), sorted(ids))

    def check_subdirs(self, base):
        self.assertEqual(
            sorted(os.listdir(base)),
            ['README.txt', 'corrupt', 'files', 'partial', 'tmp']
        )

        f = path.join(base, 'README.txt')
        self.assertFalse(path.islink(f))
        self.assertEqual(open(f, 'rt').read(), filestore.README)

        self.assertEqual(
            sorted(os.listdir(path.join(base, 'files'))),
            list(B32NAMES)
        )
        for name in ['corrupt', 'partial', 'tmp']:
            d = path.join(base, name)
            self.assertEqual(os.listdir(d), [])

        for name in ['corrupt', 'files', 'partial', 'tmp']:
            d = path.join(base, name)
            self.assertTrue(path.isdir(d))
            self.assertFalse(path.islink(d))
        for name in B32NAMES:
            d = path.join(base, 'files', name)
            self.assertTrue(path.isdir(d))
            self.assertFalse(path.islink(d))
            self.assertEqual(os.listdir(d), [])

    def test_init_dirs(self):
        tmp = TempDir()

        # Create base dir so init_dirs() doesn't get called
        base = tmp.makedirs('.dmedia')
        fs = filestore.FileStore(tmp.dir)
        self.assertEqual(os.listdir(base), [])

        # Test when no subdirs exist:
        self.assertIsNone(fs.init_dirs())
        self.check_subdirs(base)

        # Test when all subdirs exist:
        self.assertIsNone(fs.init_dirs())
        self.check_subdirs(base)

        # Test when some subdirs exist:
        os.rmdir(path.join(base, 'tmp'))
        for (i, name) in enumerate(B32NAMES):
            if i % 3 == 0:
                d = path.join(base, 'files', name)
        self.assertIsNone(fs.init_dirs())
        self.check_subdirs(base)

    def test_catch_traversal(self):
        tmp = TempDir()
        parent = tmp.makedirs('foo')
        base = tmp.join('foo', '.dmedia')
        fs = filestore.FileStore(parent)

        bad = tmp.join('foo', '.dmedia2', 'stuff')
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.catch_traversal(bad)
        self.assertEqual(cm.exception.base, base)
        self.assertEqual(cm.exception.pathname, bad)
        self.assertEqual(cm.exception.abspath, bad)

        bad = tmp.join('foo', '.dmedia', '..', '.dmedia2')
        assert '..' in bad
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.catch_traversal(bad)
        self.assertEqual(cm.exception.base, base)
        self.assertEqual(cm.exception.pathname, bad)
        self.assertEqual(cm.exception.abspath, tmp.join('foo', '.dmedia2'))

        good = tmp.join('foo', '.dmedia', 'stuff')
        self.assertEqual(fs.catch_traversal(good), good)

    def test_join(self):
        tmp = TempDir()
        parent = tmp.makedirs('foo')
        base = tmp.join('foo', '.dmedia')
        fs = filestore.FileStore(parent)

        # Test with an absolute path in parts:
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.join('files', '/root')
        self.assertEqual(cm.exception.base, base)
        self.assertEqual(cm.exception.pathname, '/root')
        self.assertEqual(cm.exception.abspath, '/root')

        # Test with some .. climbers:
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.join('NW', '..', '..', '.ssh')
        self.assertEqual(cm.exception.base, base)
        self.assertEqual(
            cm.exception.pathname,
            tmp.join('foo', '.dmedia', 'NW', '..', '..', '.ssh')
        )
        self.assertEqual(cm.exception.abspath, tmp.join('foo', '.ssh'))

        # Test for former security issue!  See:
        # https://bugs.launchpad.net/dmedia/+bug/708663
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.join('..', '.dmedia2', 'stuff')
        self.assertEqual(cm.exception.base, base)
        self.assertEqual(
            cm.exception.pathname,
            tmp.join('foo', '.dmedia', '..', '.dmedia2', 'stuff')
        )
        self.assertEqual(
            cm.exception.abspath,
            tmp.join('foo', '.dmedia2', 'stuff')
        )

        # Test with some correct parts:
        self.assertEqual(
            fs.join('NW', 'BNVXVK5DQGIOW7MYR4K3KA5K22W7NW'),
            tmp.join('foo', '.dmedia', 'NW', 'BNVXVK5DQGIOW7MYR4K3KA5K22W7NW')
        )

    def test_path(self):
        tmp = TempDir()
        parent = tmp.makedirs('foo')
        base = tmp.join('foo', '.dmedia')
        fs = filestore.FileStore(parent)

        _id = random_id()
        self.assertEqual(
            fs.path(_id),
            tmp.join('foo', '.dmedia', 'files', _id[:2], _id[2:])
        )

        # Test to make sure hashes are getting checked with safe_id():
        bad = 'NWBNVXVK5..GIOW7MYR4K3KA5K22W7NW'
        with self.assertRaises(ValueError) as cm:
            fs.path(bad)
        self.assertEqual(
            str(cm.exception),
            '_id: cannot b32decode {!r}: Non-base32 digit found'.format(bad)
        )

    def test_partial_path(self):
        tmp = TempDir()
        parent = tmp.makedirs('foo')
        base = tmp.join('foo', '.dmedia')
        fs = filestore.FileStore(parent)

        _id = random_id()
        self.assertEqual(
            fs.partial_path(_id),
            tmp.join('foo', '.dmedia', 'partial', _id)
        )

        # Test to make sure hashes are getting checked with safe_id():
        bad = 'NWBNVXVK5..GIOW7MYR4K3KA5K22W7NW'
        with self.assertRaises(ValueError) as cm:
            fs.partial_path(bad)
        self.assertEqual(
            str(cm.exception),
            '_id: cannot b32decode {!r}: Non-base32 digit found'.format(bad)
        )

    def test_corrupt_path(self):
        tmp = TempDir()
        parent = tmp.makedirs('foo')
        base = tmp.join('foo', '.dmedia')
        fs = filestore.FileStore(parent)

        _id = random_id()
        self.assertEqual(
            fs.corrupt_path(_id),
            tmp.join('foo', '.dmedia', 'corrupt', _id)
        )

        # Test to make sure hashes are getting checked with safe_id():
        bad = 'NWBNVXVK5..GIOW7MYR4K3KA5K22W7NW'
        with self.assertRaises(ValueError) as cm:
            fs.corrupt_path(bad)
        self.assertEqual(
            str(cm.exception),
            '_id: cannot b32decode {!r}: Non-base32 digit found'.format(bad)
        )

    def test_exists(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        id1 = random_id()

        # File doesn't exist
        self.assertFalse(fs.exists(id1))
        self.assertFalse(fs.exists(id1))

        # File exists
        tmp.touch('.dmedia', 'files', id1[:2], id1[2:])
        self.assertTrue(fs.exists(id1))

        # Not file:
        id2 = random_id()
        tmp.makedirs('.dmedia', 'files', id2[:2], id2[2:])
        self.assertTrue(path.isdir(fs.path(id2)))
        tmp.makedirs('.dmedia', 'files', id2[:2], id2[2:])
        self.assertTrue(path.isdir(fs.path(id2)))
        self.assertFalse(fs.exists(id2))
        self.assertFalse(fs.exists(id2))

    def test_open(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        id1 = random_id()

        # File doesn't exist
        with self.assertRaises(IOError) as cm:
            fp = fs.open(id1)
        self.assertEqual(cm.exception.errno, 2)

        # File exists
        data = b'ABC' * 1000
        canonical = fs.path(id1)
        open(canonical, 'wb').write(data)
        fp = fs.open(id1)
        self.assertIsInstance(fp, io.BufferedReader)
        self.assertEqual(fp.mode, 'rb')
        self.assertEqual(fp.tell(), 0)
        self.assertEqual(fp.name, canonical)
        self.assertEqual(fp.read(), data)

    def test_verify(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        canonical = fs.path(ID)
        corrupt = fs.corrupt_path(ID)

        # File doesn't exist
        with self.assertRaises(IOError) as cm:
            fp = fs.verify(ID)
        self.assertEqual(cm.exception.errno, 2)

        # File exists:
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.close()

        self.assertEqual(fs.verify(ID), CH)  # return_fp=False
        self.assertEqual(fs.verify(ID, False), CH)  # return_fp=False

        fp = fs.verify(ID, True)  # return_fp=True
        self.assertEqual(fp.mode, 'rb')
        self.assertEqual(fp.tell(), 0)
        self.assertEqual(fp.name, canonical)

        # File exists and is corrupted
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.write(b'F')
        fp.close()
        c = filestore.hash_fp(open(canonical, 'rb'))
        self.assertNotEqual(c.id, ID)

        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.FileIntegrityError) as cm:
            fp = fs.verify(ID)
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.parent, tmp.dir)
        self.assertEqual(cm.exception.bad_id, c.id)

        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))

    def test_verify_iter(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        canonical = fs.path(ID)
        corrupt = fs.corrupt_path(ID)

        # File doesn't exist
        with self.assertRaises(IOError) as cm:
            for (i, leaf) in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
                pass
        self.assertEqual(cm.exception.errno, 2)

        # File exists
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.close()

        j = 0
        for (i, leaf) in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
            self.assertEqual(leaf, LEAVES[i])
            self.assertEqual(i, j)
            j += 1
        self.assertEqual(i, 2)

        # File exists and file size is 1 byte too small:
        fp = open(canonical, 'wb')
        fp.write(LEAVES[0])
        fp.write(LEAVES[1])
        fp.write(LEAVES[2][:-1])
        fp.close()
        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.SizeIntegrityError) as cm:
            for (i, leaf) in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
                pass
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.parent, tmp.dir)
        self.assertEqual(cm.exception.file_size, FILE_SIZE)
        self.assertEqual(cm.exception.bad_file_size, FILE_SIZE - 1)
        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))
        os.remove(corrupt)

        # File exists and file size is 1 byte too large:
        fp = open(canonical, 'wb')
        fp.write(LEAVES[0])
        fp.write(LEAVES[1])
        fp.write(LEAVES[2] + b'c')
        fp.close()
        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.SizeIntegrityError) as cm:
            for (i, leaf) in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
                pass
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.parent, tmp.dir)
        self.assertEqual(cm.exception.file_size, FILE_SIZE)
        self.assertEqual(cm.exception.bad_file_size, FILE_SIZE + 1)
        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))
        os.remove(corrupt)

        # File exists has corruption on 2nd leaf:
        fp = open(canonical, 'wb')
        fp.write(LEAVES[0])
        fp.write(LEAVES[1][:-1] + b'f')
        fp.write(LEAVES[2])
        fp.close()
        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.LeafIntegrityError) as cm:
            for (i, leaf) in fs.verify_iter(ID, FILE_SIZE, LEAF_HASHES):
                pass
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.parent, tmp.dir)
        self.assertEqual(cm.exception.leaf_index, 1)
        self.assertEqual(
            cm.exception.leaf_hash,
            filestore.hash_leaf(1, LEAVES[1])
        )
        self.assertEqual(
            cm.exception.bad_leaf_hash,
            filestore.hash_leaf(1, LEAVES[1][:-1] + b'f')
        )
        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))
        os.remove(corrupt)

    def test_verify_iter2(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        canonical = fs.path(ID)
        corrupt = fs.corrupt_path(ID)

        # File doesn't exist
        with self.assertRaises(IOError) as cm:
            leaves = tuple(fs.verify_iter2(ID))
        self.assertEqual(cm.exception.errno, 2)

        # File exists:
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.close()

        self.assertEqual(
            tuple(fs.verify_iter2(ID)),
            tuple(enumerate(LEAVES))
        )

        # File exists and is corrupted
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.write(b'F')
        fp.close()
        c = filestore.hash_fp(open(canonical, 'rb'))
        self.assertNotEqual(c.id, ID)

        self.assertTrue(path.isfile(canonical))
        self.assertFalse(path.exists(corrupt))

        with self.assertRaises(filestore.FileIntegrityError) as cm:
            leaves = tuple(fs.verify_iter2(ID))
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.parent, tmp.dir)
        self.assertEqual(cm.exception.bad_id, c.id)

        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(corrupt))

    def test_remove(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        canonical = fs.path(_id)

        # File doesn't exist
        with self.assertRaises(OSError) as cm:
            fs.remove(_id)
        self.assertEqual(cm.exception.errno, 2)

        # File exists
        open(canonical, 'wb').write(b'hello')
        self.assertTrue(path.isfile(canonical))
        fs.remove(_id)
        self.assertFalse(path.exists(canonical))

    def test_allocate_tmp(self):
        tmp = TempDir()
        t = tmp.join('.dmedia', 'tmp')
        fs = filestore.FileStore(tmp.dir)
        self.assertTrue(path.isdir(t))

        # Test with size=2311
        fp = fs.allocate_tmp(2311)
        self.assertIsInstance(fp, io.BufferedRandom)
        self.assertEqual(fp.mode, 'rb+')  # FIXME: Hmmm, 'r+b' becomes 'rb+' ?
        stat = os.fstat(fp.fileno())
        self.assertTrue(stat.st_size in [0, 2311])
        self.assertEqual(path.dirname(fp.name), t)
        self.assertTrue(
            '.' not in path.basename(fp.name)
        )

    def test_allocate_partial(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        filename = tmp.join('.dmedia', 'partial', _id)

        # Test when file dosen't yet exist
        fp = fs.allocate_partial(2311, _id)
        self.assertIsInstance(fp, io.IOBase)
        self.assertEqual(fp.name, filename)
        stat = os.fstat(fp.fileno())

        self.assertTrue(fp.mode in ['rb+', 'wb'])
        if fp.mode == 'r+b':
            self.assertTrue(stat.st_size in (0, 2311))
        if fp.mode == 'wb':
            self.assertEqual(stat.st_size, 0)
        fp.write(b'a' * 3141)  # Write something to file
        fp.close()

        # Test with pre-existing file longer than size:
        fp = fs.allocate_partial(2311, _id)
        self.assertIsInstance(fp, io.IOBase)
        self.assertEqual(fp.name, filename)
        stat = os.fstat(fp.fileno())

        self.assertEqual(fp.mode, 'rb+')
        self.assertEqual(stat.st_size, 2311)  # fp.trucate() was called

    def test_move_to_canonical(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        dst = fs.path(_id)

        # Test with wrong src_fp type
        with self.assertRaises(TypeError) as cm:
            fs.move_to_canonical(17, _id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', io.BufferedIOBase, int, 17)
        )
        self.assertFalse(path.exists(dst))

        # Test with src_fp opened in wrong mode
        bad = tmp.touch('.dmedia', 'bad1.mov')
        for mode in ('r', 'w', 'r+'):
            fp = open(bad, mode)
            with self.assertRaises(TypeError) as cm:
                fs.move_to_canonical(fp, _id)
            self.assertEqual(
                str(cm.exception),
                TYPE_ERROR.format('src_fp', io.BufferedIOBase, type(fp), fp)
            )
            self.assertFalse(path.exists(dst))
            self.assertFalse(fp.closed)

        # Test when filename isn't contained in base (to ensure that
        # FileStore.catch_traversal() is used to validate src_fp.name)
        bad = tmp.touch('.dmedia', '..', 'bad2.mov')
        fp = open(bad, 'rb')
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.move_to_canonical(fp, _id)
        self.assertEqual(cm.exception.abspath, tmp.join('bad2.mov'))
        self.assertEqual(cm.exception.base, fs.base)
        self.assertFalse(path.exists(dst))
        self.assertFalse(fp.closed)

        # Test that _id is validated with safe_id()
        good = tmp.write(b'yup', '.dmedia', 'tmp', 'good.mov')
        os.chmod(good, 0o660)
        fp = open(good, 'rb')
        b32 = 'NWBNVXVK5DQGIOW7MYR4K3KA'
        with self.assertRaises(ValueError) as cm:
            fs.move_to_canonical(fp, b32)
        self.assertEqual(
            str(cm.exception),
            "_id: decoded must be 30 bytes; got 15: 'NWBNVXVK5DQGIOW7MYR4K3KA'"
        )
        self.assertFalse(path.exists(dst))
        self.assertFalse(fp.closed)

        # Test when src_fp is closed
        fp.close()
        with self.assertRaises(ValueError) as cm:
            fs.move_to_canonical(fp, _id)
        self.assertEqual(
            str(cm.exception),
            'src_fp is closed, must be open: {!r}'.format(fp)
        )
        self.assertFalse(path.exists(dst))
        self.assertTrue(fp.closed)

        # Test when it's all good
        fp = open(good, 'rb')
        self.assertEqual(stat.S_IMODE(os.stat(good).st_mode), 0o660)
        self.assertEqual(fs.move_to_canonical(fp, _id), dst)
        self.assertTrue(fp.closed)
        self.assertFalse(path.exists(good))
        self.assertTrue(path.isfile(dst))
        self.assertEqual(stat.S_IMODE(os.stat(dst).st_mode), 0o444)
        self.assertEqual(open(dst, 'rb').read(), b'yup')

    def test_move_to_corrupt(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)
        _id = random_id()
        corrupt = fs.corrupt_path(_id)
        canonical = fs.path(_id)

        # Test with wrong fp type
        with self.assertRaises(TypeError) as cm:
            fs.move_to_corrupt(17, _id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', io.BufferedReader, int, 17)
        )
        self.assertFalse(path.exists(corrupt))

        # Test with fp opened 'r' mode:
        bad = tmp.touch('.dmedia', 'bad1.mov')
        fp = open(bad, 'r')
        with self.assertRaises(TypeError) as cm:
            fs.move_to_corrupt(fp, _id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', io.BufferedReader, io.TextIOWrapper, fp)
        )
        self.assertFalse(path.exists(corrupt))
        self.assertFalse(fp.closed)

        # Test with fp opened 'rb+' mode:
        fp = open(bad, 'rb+')
        with self.assertRaises(TypeError) as cm:
            fs.move_to_corrupt(fp, _id)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('src_fp', io.BufferedReader, io.BufferedRandom, fp)
        )
        self.assertFalse(path.exists(corrupt))
        self.assertFalse(fp.closed)

        # Test when filename isn't contained in base (to ensure that
        # FileStore.catch_traversal() is used to validate tmp_fp.name)
        bad = tmp.touch('.dmedia', '..', 'bad2.mov')
        fp = open(bad, 'rb')
        with self.assertRaises(filestore.PathTraversal) as cm:
            fs.move_to_corrupt(fp, _id)
        self.assertEqual(cm.exception.abspath, tmp.join('bad2.mov'))
        self.assertEqual(cm.exception.base, fs.base)
        self.assertFalse(path.exists(corrupt))
        self.assertFalse(fp.closed)

        # Test that _id is validated with safe_id()
        good = tmp.write(b'yup', '.dmedia', 'tmp', 'good.mov')
        fp = open(good, 'rb')
        b32 = 'NWBNVXVK5DQGIOW7MYR4K3KA'
        with self.assertRaises(ValueError) as cm:
            fs.move_to_corrupt(fp, b32)
        self.assertEqual(
            str(cm.exception),
            "_id: decoded must be 30 bytes; got 15: 'NWBNVXVK5DQGIOW7MYR4K3KA'"
        )
        self.assertFalse(path.exists(corrupt))
        self.assertFalse(fp.closed)

        # Test when fp is closed
        fp.close()
        with self.assertRaises(ValueError) as cm:
            fs.move_to_corrupt(fp, _id)
        self.assertEqual(
            str(cm.exception),
            'src_fp is closed, must be open: {!r}'.format(fp)
        )
        self.assertFalse(path.exists(corrupt))
        self.assertTrue(fp.closed)

        # Test when it's all good
        fp = open(good, 'rb')
        self.assertEqual(fs.move_to_corrupt(fp, _id), corrupt)
        self.assertTrue(fp.closed)
        self.assertFalse(path.exists(good))
        self.assertTrue(path.isfile(corrupt))
        self.assertEqual(open(corrupt, 'rb').read(), b'yup')

    def test_import_file(self):
        tmp = TempDir()
        src = tmp.join('movie.mov')
        write_sample_file(src)

        fs = filestore.FileStore(tmp.dir)
        dst = fs.path(ID)
        self.assertTrue(path.isfile(src))
        self.assertFalse(path.exists(dst))

        src_fp = open(src, 'rb')
        self.assertEqual(fs.import_file(src_fp), CH)
        self.assertTrue(path.isfile(src))
        self.assertTrue(path.isfile(dst))

        self.assertEqual(filestore.hash_fp(open(dst, 'rb')), CH)

    def test_copy(self):
        tmp = TempDir()
        fs = filestore.FileStore(tmp.dir)

        tmp1 = TempDir()
        tmp2 = TempDir()
        fs1 = filestore.FileStore(tmp1.dir)
        fs2 = filestore.FileStore(tmp2.dir)

        with self.assertRaises(TypeError) as cm:
            fs.copy(ID)
        self.assertEqual(
            str(cm.exception),
            'FileStore.copy() takes at least one destination filestore'
        )

        # Bad type in filestores
        with self.assertRaises(TypeError) as cm:
            fs.copy(ID, 17, fs2)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('filestores[0]', filestore.FileStore, int, 17)
        )
        with self.assertRaises(TypeError) as cm:
            fs.copy(ID, fs1, 18.0)
        self.assertEqual(
            str(cm.exception),
            TYPE_ERROR.format('filestores[1]', filestore.FileStore, float, 18.0)
        )

        # Same parent in filestores
        fs_same = filestore.FileStore(tmp.dir)
        with self.assertRaises(ValueError) as cm:
            fs.copy(ID, fs_same, fs2)
        self.assertEqual(
            str(cm.exception),
            'wont copy into filestores[0], same parent as {!r}'.format(fs)
        )
        with self.assertRaises(ValueError) as cm:
            fs.copy(ID, fs1, fs_same)
        self.assertEqual(
            str(cm.exception),
            'wont copy into filestores[1], same parent as {!r}'.format(fs)
        )

        # src file doesn't exist:
        with self.assertRaises(IOError) as cm:
            fs.copy(ID, fs1, fs2)
        self.assertEqual(cm.exception.errno, 2)

        # File exists and is corrupted
        canonical = fs.path(ID)
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.write(b'F')
        fp.close()
        c = filestore.hash_fp(open(canonical, 'rb'))
        self.assertNotEqual(c.id, ID)

        with self.assertRaises(filestore.FileIntegrityError) as cm:
            fs.copy(ID, fs1, fs2)
        self.assertEqual(cm.exception.parent, fs.parent)
        self.assertEqual(cm.exception.id, ID)
        self.assertEqual(cm.exception.bad_id, c.id)
        self.assertFalse(path.exists(canonical))
        self.assertTrue(path.isfile(fs.corrupt_path(ID)))
        os.remove(fs.corrupt_path(ID))

        self.assertFalse(fs1.exists(ID))
        self.assertEqual(os.listdir(fs1.tmp), [])
        self.assertEqual(list(fs1), [])

        self.assertFalse(fs2.exists(ID))
        self.assertEqual(os.listdir(fs2.tmp), [])
        self.assertEqual(list(fs2), [])


        # File exists and is valid
        fp = open(canonical, 'wb')
        for leaf in LEAVES:
            fp.write(leaf)
        fp.close()
        self.assertEqual(fs.copy(ID, fs1, fs2), CH)

        self.assertTrue(fs1.exists(ID))
        self.assertEqual(list(fs1), [ID])
        self.assertEqual(os.listdir(fs1.tmp), [])

        self.assertTrue(fs2.exists(ID))
        self.assertEqual(list(fs2), [ID])
        self.assertEqual(os.listdir(fs2.tmp), [])

        fs1.verify(ID)
        fs2.verify(ID)
        fs.verify(ID)


        # File already exists in destination filestores
        self.assertEqual(fs.copy(ID, fs1, fs2), CH)

        self.assertTrue(fs1.exists(ID))
        self.assertEqual(list(fs1), [ID])
        self.assertEqual(os.listdir(fs1.tmp), [])

        self.assertTrue(fs2.exists(ID))
        self.assertEqual(list(fs2), [ID])
        self.assertEqual(os.listdir(fs2.tmp), [])

        fs1.verify(ID)
        fs2.verify(ID)
        fs.verify(ID)
